"""
JSON encoders and loaders using the jsonyx library.
See https://github.com/nineteendo/jsonyx
"""

from typing import Any


def jsonyx_encoder(obj: Any) -> str:
    """JSONYX encoder which allows non-string dict keys.
    Here it runs in very verbose mode."""
    import jsonyx.allow

    return jsonyx.dumps(
        obj,
        sort_keys=True,
        indent=2,
        allow=jsonyx.allow.EVERYTHING,
    )


def jsonyx_compactish_encoder(obj: Any) -> str:
    """JSONYX encoder which allows non-string dict keys.
    Here it runs in a semi-compact mode where arrays and dicts containing only primitives are not indented."""
    import jsonyx.allow

    return jsonyx.dumps(
        obj,
        sort_keys=True,
        indent=2,
        indent_leaves=False,
        allow=jsonyx.allow.EVERYTHING,
    )


def jsonyx_compact_encoder(obj: Any) -> str:
    """JSONYX encoder which allows non-string dict keys.
    Here it runs in very compact mode."""
    import jsonyx.allow

    return jsonyx.dumps(
        obj,
        sort_keys=True,
        allow=jsonyx.allow.EVERYTHING,
    )


def jsonyx_permissive_loader(text: str) -> Any:
    """JSONYX loader in very permissive mode."""
    import jsonyx.allow

    return jsonyx.loads(text, allow=jsonyx.allow.EVERYTHING)



import pytest

from pytest_respect.resources import TestResources


@pytest.fixture
def resources(request: pytest.FixtureRequest) -> TestResources:
    """Load file resources relative to test functions and fixtures."""
    accept = request.config.getoption("--respect-accept-max")
    resources = TestResources(request, accept_count=accept)
    resources.default.ndigits = 4
    return resources


def pytest_addoption(parser):
    group = parser.getgroup("respect")
    group.addoption(
        "--respect-accept",
        action="store_const",
        dest="respect_accept_max",
        const=1e9,
        help="When results don't match expectations, create or update the expected files instead of failing the tests.",
    )
    group.addoption(
        "--respect-accept-one",
        action="store_const",
        dest="respect_accept_max",
        const=1,
        help="Like --respect-accept but only for one expectation in each test.",
    )
    group.addoption(
        "--respect-accept-max",
        default=0,
        type=int,
        help="Like --respect-accept but only for this many expectations in each test.",
    )



"""Fixture for loading resources relative to test functions and fixtures."""

import builtins
import fnmatch
import inspect
import json
from collections.abc import Callable, Iterable, Sequence
from pathlib import Path
from types import EllipsisType, UnionType
from typing import Any, Protocol, TypeVar

from pytest import FixtureRequest

# Optional imports falling back to stub implementations to make the type checker happy
try:
    from pydantic import BaseModel, TypeAdapter
except ImportError:  # pragma: no cover
    from ._fakes import BaseModel, TypeAdapter


# Dont' include in pytest tracebacks. We patch this out in unit tests to see where in our code errors occur.
__tracebackhide__ = True

from pytest_respect.utils import coalesce, prepare_for_json_encode

DEFAULT_RESOURCES_DIR = "resources"

PMT = TypeVar("PMT", bound=BaseModel)
T = TypeVar("T")


PathParts = tuple[Path, str | None]
"""Directory and base file-name for a resource path"""


class PathMaker(Protocol):
    """Protocol for functions which determine a directory and base file-name from the coordinates of a test function."""

    def __call__(
        self,
        test_dir: Path,
        test_file_name: str,
        test_class_name: str | None,
        test_name: str,
    ) -> PathParts: ...


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Static Resource Listing


def list_dir(
    dir: Path,
    include: str | Sequence[str] = "*",
    *,
    exclude: str | Sequence[str] = tuple(),
    strip_ext: str | bool = False,
) -> list[str]:
    """List the names of the files in the directory, including and excluding globs."""
    if isinstance(include, str):
        include = [include]
    if isinstance(exclude, str):
        exclude = [exclude]

    names: set[str] = set()
    for inc in include:
        names.update(path.name for path in dir.glob(inc))
    for exc in exclude:
        names = {name for name in names if not fnmatch.fnmatch(name, exc)}
    name_list = sorted(names)
    name_list = strip_extensions(name_list, strip_ext=strip_ext)
    return name_list


def list_resources(
    include: str | Sequence[str] = "*",
    *,
    exclude: str | Sequence[str] = tuple(),
    path_maker: PathMaker | None = None,
    strip_ext: str | bool = False,
) -> builtins.list[str]:
    """Static version of TestResources.list method which can be used for parametric tests.

    This version does not have access to a FixtureRequest object so we use the directory of the calling module instead.
    We have no test class or test function to pass to the path_maker and we use only the directory from its return
    value.

    Parametric Test Example:

        >>> @pytest.mark.parametrize("curve_name", list_resources("power_curve_*.json")
        >>> def test_power_curve(resources, curve_name):
        >>>    curve = resources.load_pydantic(PowerCurve, curve_name)
        >>>    assert len(curve.points) >= 2

    Parametric Fixture Example:

        >>> @pytest.fixture(params=list_resources("power_curve_*.json", exclude=["*__actual*"], strip_ext=".json"))
        >>> def each_power_curve(request, resources) -> PowerCurve:
        >>>    return resources.load_pydantic(PowerCurve, request.param)

    Args:

        include: one or more glob patterns for filenames to include
        exclude: zero or more glob patterns for filenames to exclude
        path_maker: Function to turn test co-ordinates into a directory to look for
            resources in. The default uses the test file name for the directory.
        strip_ext: Whether to strip extension from listed resource file names. If
            True or False then strip all or no extensions from last dot.
            If string, then strip only that string from the ends of file names. A
            strip_ext string must include the dot (if wanted) to allow stripping
            suffixes such as `__bad.json`

    """
    if path_maker is None:
        path_maker = TestResources.pm_only_file

    calling_frame = inspect.stack()[1]
    test_file = Path(calling_frame.filename)

    dir_path, _ = path_maker(
        test_dir=test_file.parent,
        test_file_name=test_file.name.rsplit(".", 1)[0],
        test_class_name=None,
        test_name="",
    )

    resources = list_dir(dir_path, include, exclude=exclude)
    resources = strip_extensions(resources, strip_ext)
    return resources


def strip_extensions(
    resources: Iterable[str],
    strip_ext: str | bool = False,
) -> list[str]:
    """Strip file extensions from a list of resources, depending on strip_ext.

    Args:
        resources: list of resources to maybe strip extensions from.
        strip_ext: Whether to strip extension from listed resource file names. If
            True or False then strip all or no extensions from last dot.
            If string, then strip only that string from the ends of file names. A
            strip_ext string must include the dot (if wanted) to allow stripping
            suffixes such as `__bad.json`

    Returns:
        The original resources or a copy with extensions stripped.

    """
    if strip_ext is True:
        resources = [r.rsplit(".", 1)[0] for r in resources]
    elif isinstance(strip_ext, str):
        resources = [r.removesuffix(strip_ext) for r in resources]
    elif not isinstance(strip_ext, list):
        resources = list(resources)
    return resources


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# JSON encoders & Decoders

JsonEncoder = Callable[[Any], str]
"""Function to convert data to JSON encoded text."""

JsonLoader = Callable[[str], Any]
"""Function to convert JSON encoded text to python data."""


def python_json_encoder(obj: Any) -> str:
    """Standard JSON encoder in very verbose mode."""
    return json.dumps(obj, sort_keys=True, indent=2)


def python_compact_json_encoder(obj: Any) -> str:
    """Standard JSON encoder in very compact mode."""
    return json.dumps(obj, sort_keys=True, indent=None)


def python_json_loader(text: str) -> Any:
    """Standard JSON loader."""
    return json.loads(text)


class Defaults:
    def __init__(self):
        self.ndigits: int | None = None
        """How many digits to round floats to by default when comparing JSON data."""

        self.json_encoder: JsonEncoder = python_json_encoder
        """Function used to convert data to JSON encoded text. Defaults to standard python JSON encoding."""

        self.json_loader: JsonLoader = python_json_loader
        """Function used to convert JSON encoded text to python data. Defaults to standard python JSON decoding."""

        self.path_maker: PathMaker = TestResources.pm_class
        """
        Function used to make paths to resources. Defaults to as pm_class making resource paths like
        ``<dir>/test_file__TestClass/test_method.<ext>``, omitting the ``__TestClass`` part if we are not in a class.
        """

    def _ndigits(self, value: int | None | EllipsisType) -> int | None:
        """Resolve ndigits given the default value here and optional override."""
        return coalesce(self.ndigits, value, nonable=True)

    def _json_encoder(self, value: JsonEncoder | None | EllipsisType) -> JsonEncoder:
        """Resolve json_encoder given the default value here and optional override."""
        return coalesce(self.json_encoder, value)

    def _json_loader(self, value: JsonLoader | None | EllipsisType) -> JsonLoader:
        """Resolve json_loader given the default value here and optional override."""
        return coalesce(self.json_loader, value)

    def _path_maker(self, value: PathMaker | None | EllipsisType) -> PathMaker:
        """Resolve path_maker given the default value here and optional override."""
        return coalesce(self.path_maker, value)


class TestResources:
    __test__ = False  # Don't try to collect this as a test

    def __init__(self, request: FixtureRequest, accept_count: int = 0):
        """Create test resources instance, usually in a function-scoped fixture.

        Args:
            request: The pytest fixture request object.
            ndigits: How many digits to round floats to by default when comparing JSON data and objects which are
                converted to JSON before comparison. Defaults to no rounding.
            accept_count: Accept the actual results for this many mismatches before failing the test.

        """
        self.request: FixtureRequest = request
        """The pytest fixture request that we get context information from."""

        self.accept_count: int = accept_count
        """Mimatches remaining before we fail the test."""

        self.default: Defaults = Defaults()
        """Default behaviours. They can be optionally overridden in methods that use them."""

        self._extra_preppers: list[tuple[type | UnionType, Callable[[Any], Any]]] = []
        """JSON Preppers to apply after the global JSON_PREPPERS."""

    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
    # Path Makers

    @staticmethod
    def pm_function(test_dir: Path, test_file_name: str, test_class_name: str | None, test_name: str) -> PathParts:
        """PathMaker to build directory from test_file, class if present, and function. No contribution is made to the
        file name.

        - ``<dir>/test_file__TestClass__test_method/data.<ext>``
        - ``<dir>/test_file__test_method/data.<ext>``
        """
        if test_class_name:
            dir = test_dir / f"{test_file_name}__{test_class_name}__{test_name}"
        else:
            dir = test_dir / f"{test_file_name}__{test_name}"

        return dir, None

    @staticmethod
    def pm_class(test_dir: Path, test_file_name: str, test_class_name: str | None, test_name: str) -> PathParts:
        """PathMaker to build directory from test_file and class if present, and file-name from test method.

        This is the default method for constructing resource paths.

        - ``<dir>/test_file__TestClass/test_method.<ext>``
        - ``<dir>/test_file/test_method.<ext>``
        """
        if test_class_name:
            dir = test_dir / f"{test_file_name}__{test_class_name}"
        else:
            dir = test_dir / test_file_name
        return dir, test_name

    @staticmethod
    def pm_only_class(test_dir: Path, test_file_name: str, test_class_name: str | None, test_name: str) -> PathParts:
        """PathMaker to build directory from test_file and class if present and contribute nothing to the file-name

        - ``<dir>/test_file__TestClass/data.<ext>``
        - ``<dir>/test_file/data.<ext>``
        """
        if test_class_name:
            dir = test_dir / f"{test_file_name}__{test_class_name}"
        else:
            dir = test_dir / test_file_name
        return dir, None

    @staticmethod
    def pm_file(test_dir: Path, test_file_name: str, test_class_name: str | None, test_name: str) -> PathParts:
        """PathMaker to build directory from test_file and file-name from test class if present, and test method.

        - ``<dir>/test_file__TestClass/test_method.<ext>``
        - ``<dir>/test_file/test_method.<ext>``
        """
        if test_class_name:
            file = f"{test_class_name}__{test_name}"
        else:
            file = test_name
        return test_dir / test_file_name, file

    @staticmethod
    def pm_only_file(
        test_dir: Path,
        test_file_name: str,
        test_class_name: str | None = None,
        test_name: str | None = None,
    ) -> tuple[Path, str | None]:
        """PathMaker to build directory from test_file and contribute to the file-name.

        - ``<dir>/test_file/data.<ext>``
        """
        return test_dir / test_file_name, None

    @staticmethod
    def pm_dir(test_dir: Path, test_file_name: str, test_class_name: str | None, test_name: str) -> PathParts:
        """PathMaker to use "resources" for directory and build file-name from test file, test class if present and
        test function.

        - ``<dir>/resources/test_file__TestClass__test_method.<ext>``
        - ``<dir>/resources/test_file__test_method.<ext>``
        """
        path_maker = TestResources.pm_dir_named(DEFAULT_RESOURCES_DIR)
        return path_maker(test_dir, test_file_name, test_class_name, test_name)

    @staticmethod
    def pm_dir_named(dir_name: str = DEFAULT_RESOURCES_DIR) -> PathMaker:
        """PathMaker to use the given name for directory and build file-name from test file, test class if present and
        test function.

        - ``<dir>/<dir_name>/test_file__TestClass__test_method.<ext>``
        - ``<dir>/<dir_name>/test_file__test_method.<ext>``
        """

        def path_from_dir(
            test_dir: Path,
            test_file_name: str,
            test_class_name: str | None,
            test_name: str,
        ) -> PathParts:
            if test_class_name:
                file = f"{test_file_name}__{test_class_name}__{test_name}"
            else:
                file = f"{test_file_name}__{test_name}"
            return test_dir / dir_name, file

        return path_from_dir

    @staticmethod
    def pm_only_dir(
        test_dir: Path,
        test_file_name: str,
        test_class_name: str | None = None,
        test_name: str | None = None,
    ) -> PathParts:
        """Use "resources" for directory and nothing for the file-name."""
        return test_dir / DEFAULT_RESOURCES_DIR, None

    @staticmethod
    def pm_only_dir_named(dir_name: str = DEFAULT_RESOURCES_DIR) -> PathMaker:
        """Use the given name for directory nothing for the file-name."""

        def path_from_dir(
            test_dir: Path,
            test_file_name: str,
            test_class_name: str | None,
            test_name: str,
        ) -> PathParts:
            return test_dir / dir_name, None

        return path_from_dir

    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
    # Paths

    def dir(self, path_maker: PathMaker | None = None) -> Path:
        """Directory for resources belonging to the test.

        Args:
            path_maker: Function to turn test co-ordinates into a directory and partial
                file-name for the resource. The file part will be ignored.

        """
        path_maker = self.default._path_maker(path_maker)
        test_file = Path(self.request.node.fspath)
        test_class: type | None = self.request.node.cls

        dir_path, base_name = path_maker(
            test_dir=test_file.parent,
            test_file_name=test_file.name.rsplit(".", 1)[0],
            test_class_name=test_class.__name__ if test_class else None,
            test_name=self.request.node.originalname,
        )

        return dir_path

    def path(
        self,
        *parts: Any,
        ext: str | None = None,
        path_maker: PathMaker | None = None,
    ) -> Path:
        """Path to a resource file within this test file/class's directory.

        Args:
            parts: Concatenate these parts to the function name, with __ separator to
                make file-name.
            ext: Add this extension to the file-name if not already present. Defaults
                to nothing which can be appropriate if we knot that the last ``part``
                already has an extension.
            path_maker: Function to turn test co-ordinates into a directory and partial
                file-name for the resource. Defaults to one which uses the test file
                (and class if present) for the dir and the test function for the file.

        """
        path_maker = self.default._path_maker(path_maker)
        test_file = Path(self.request.node.fspath)
        test_class: type | None = self.request.node.cls

        dir_path, base_name = path_maker(
            test_dir=test_file.parent,
            test_file_name=test_file.name.rsplit(".", 1)[0],
            test_class_name=test_class.__name__ if test_class else None,
            test_name=self.request.node.originalname,
        )

        name_parts: list[str] = []
        if base_name:
            name_parts.append(base_name)
        name_parts.extend(map(str, parts))
        if not name_parts:
            # If we have not file-name parts from anywhere, then fall back to "data"
            name_parts = ["data"]

        name = "__".join(name_parts)
        if ext and not name.endswith(f".{ext}"):
            name = f"{name}.{ext}"
        return dir_path / name

    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
    # List Resources

    def list(
        self,
        include: str | Sequence[str] = "*",
        *,
        exclude: str | Sequence[str] = tuple(),
        path_maker: PathMaker | None = None,
        strip_ext: str | bool = False,
    ) -> list[str]:
        """List the names of the resources in the folder.

        Args:
            include: one or more glob patterns for filenames to include
            exclude: zero or more glob patterns for filenames to exclude
            path_maker: Function to turn test co-ordinates into a directory and partial
                file-name for the resource. The file part will be ignored.
        strip_ext: Whether to strip extension from listed resource file names. If
            True or False then strip all or no extensions from last dot.
            If string, then strip only that string from the ends of file names. A
            strip_ext string must include the dot (if wanted) to allow stripping
            suffixes such as `__bad.json`

        """
        dir: Path = self.dir(path_maker)
        return list_dir(dir, include, exclude=exclude, strip_ext=strip_ext)

    def delete(
        self,
        *parts,
        ext: str | None = None,
        path_maker: PathMaker | None = None,
    ):
        """Delete a string resource relative to the current test."""
        path = self.path(*parts, ext=ext, path_maker=path_maker)
        path.unlink(missing_ok=True)

    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
    # Text Resources

    def load_text(
        self,
        *parts,
        ext: str = "txt",
        path_maker: PathMaker | None = None,
    ) -> str:
        """Load a string resource relative to the current test."""
        path = self.path(*parts, ext=ext, path_maker=path_maker)
        return path.read_text()

    def save_text(
        self,
        text: str,
        *parts,
        ext: str = "txt",
        path_maker: PathMaker | None = None,
    ) -> None:
        """Write a string to a resource relative to the current test."""
        path = self.path(*parts, ext=ext, path_maker=path_maker)
        if not path.parent.exists():
            path.parent.mkdir(parents=True, exist_ok=True)
        print(f"write {len(text)} chars to {path}")
        path.parent.mkdir(parents=False, exist_ok=True)
        path.write_text(text)

    def delete_text(
        self,
        *parts,
        ext: str = "txt",
        path_maker: PathMaker | None = None,
    ):
        """Delete a json resource relative to the current test."""
        self.delete(*parts, ext=ext, path_maker=path_maker)

    def expect_text(
        self,
        actual: str,
        *parts,
        ext: str = "txt",
        path_maker: PathMaker | None = None,
    ) -> None:
        """Assert that the actual value matches the content from resource."""
        expected_path = self.path(*parts, ext=ext, path_maker=path_maker)
        actual_path = self.path(*parts, "actual", ext=ext, path_maker=path_maker)

        if not expected_path.is_file():
            expected_path.parent.mkdir(parents=False, exist_ok=True)
            if self._accept_one():
                expected_path.write_text(actual)
                print(f"A new expectation file was written to {expected_path}.")
                actual_path.unlink(missing_ok=True)
            else:
                actual_path.write_text(actual)
                raise AssertionError(
                    f"The expectation file was not found at {expected_path}. The actual value has been written to {actual_path}."
                )

        expected = expected_path.read_text()

        try:
            assert actual == expected, (
                f"The actual value did not match the content of {expected_path}. It has been written to {actual_path} for comparison."
            )
            if actual_path.exists():
                actual_path.unlink(missing_ok=True)
                print("removing matching actual file", actual_path)
        except AssertionError:
            if self._accept_one():
                print(f"The expectation file was updated at {expected_path}.")
                expected_path.write_text(actual)
                actual_path.unlink(missing_ok=True)
            else:
                actual_path.write_text(actual)
                raise

    def _accept_one(self) -> bool:
        """Whether to accept one more mismatch before failing the test."""
        self.accept_count -= 1
        return self.accept_count >= 0

    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
    # JSON Resources

    def add_json_prepper(self, type_: type | UnionType, prepper: Callable[[Any], Any]) -> None:
        """Register a local JSON prepper for a given type, including sub-classes.

        The prepper can return a few kinds of values:
        - Simple value: encoded as-is and must be JSON serializable.
        - Dict: encoded recursively but must have keys that are supported by the json_encoder in use, usually str.
        - Collection: list, tuple, set, etc will be recursively encoded as a list.

        It can also raise AbortJsonPrep to skip this prepper and continue trying others.
        """
        self._extra_preppers.append((type_, prepper))

    def load_json(
        self,
        *parts,
        ext: str = "json",
        path_maker: PathMaker | None = None,
        json_loader: JsonLoader | EllipsisType = ...,
    ) -> Any:
        """Load a json resource relative to the current test."""
        path = self.path(*parts, ext=ext, path_maker=path_maker)
        try:
            text = path.read_text()
            json_loader = self.default._json_loader(json_loader)
            return json_loader(text)
        except Exception as e:
            raise ValueError(f"Failed to load JSON resource {path}: {repr(e)}") from e

    def data_to_json(
        self,
        data: Any,
        json_encoder: JsonEncoder | EllipsisType = ...,
        ndigits: int | None | EllipsisType = ...,
    ) -> str:
        """Convert data to a json string used for both expectations and save_json."""
        ndigits = self.default._ndigits(ndigits)
        data = prepare_for_json_encode(data, ndigits=ndigits, extra_preppers=self._extra_preppers)
        json_encoder = self.default._json_encoder(json_encoder)
        text = json_encoder(data)
        if not text.endswith("\n"):
            text += "\n"
        return text

    def save_json(
        self,
        data: Any,
        *parts,
        ext: str = "json",
        path_maker: PathMaker | None = None,
        json_encoder: JsonEncoder | EllipsisType = ...,
        ndigits: int | None | EllipsisType = ...,
    ) -> None:
        """Write JSON data to a resource relative to the current test."""
        text = self.data_to_json(data, json_encoder=json_encoder, ndigits=ndigits)
        self.save_text(text, *parts, ext=ext, path_maker=path_maker)

    def delete_json(
        self,
        *parts,
        ext: str = "json",
        path_maker: PathMaker | None = None,
    ):
        """Delete a json resource relative to the current test."""
        self.delete(*parts, ext=ext, path_maker=path_maker)

    def expect_json(
        self,
        actual: Any,
        *parts,
        ext: str = "json",
        path_maker: PathMaker | None = None,
        json_encoder: JsonEncoder | EllipsisType = ...,
        ndigits: int | None | EllipsisType = ...,
    ) -> None:
        """Assert that the actual value encodes to the JSON content from resource."""
        actual_text = self.data_to_json(actual, json_encoder=json_encoder, ndigits=ndigits)
        self.expect_text(
            actual_text,
            *parts,
            ext=ext,
            path_maker=path_maker,
        )

    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
    # Pydantic Resources

    def load_pydantic(
        self,
        model_class: type[PMT],
        *parts,
        ext: str = "json",
        path_maker: PathMaker | None = None,
        json_loader: JsonLoader | EllipsisType = ...,
    ) -> PMT:
        """Load a pydantic resource relative to the current test."""
        data = self.load_json(*parts, ext=ext, path_maker=path_maker, json_loader=json_loader)
        return model_class.model_validate(data)

    def save_pydantic(
        self,
        data: Any,
        *parts,
        ext: str = "json",
        path_maker: PathMaker | None = None,
        json_encoder: JsonEncoder | EllipsisType = ...,
        ndigits: int | None | EllipsisType = ...,
        context: Any = None,
    ) -> None:
        """Write pydantic data to a resource relative to the current test."""
        actual_data = data.model_dump(mode="json", context=context)
        self.save_json(actual_data, *parts, ext=ext, path_maker=path_maker, json_encoder=json_encoder, ndigits=ndigits)

    def delete_pydantic(self, *parts, ext: str = "json", path_maker: PathMaker | None = None):
        """Delete a json resource relative to the current test."""
        self.delete(*parts, ext=ext, path_maker=path_maker)

    def expect_pydantic(
        self,
        actual: BaseModel,
        *parts,
        ext: str = "json",
        path_maker: PathMaker | None = None,
        json_encoder: JsonEncoder | EllipsisType = ...,
        ndigits: int | None | EllipsisType = ...,
        context: Any = None,
    ) -> None:
        """Assert that the actual value encodes to the JSON content from resource."""
        actual_data = actual.model_dump(mode="json", context=context)
        self.expect_json(
            actual_data,
            *parts,
            ext=ext,
            path_maker=path_maker,
            json_encoder=json_encoder,
            ndigits=ndigits,
        )

    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
    # Pydantic TypeAdapter Resources

    def load_pydantic_adapter(
        self,
        type_: type[T] | TypeAdapter[T],
        *parts,
        ext: str = "json",
        path_maker: PathMaker | None = None,
        json_loader: JsonLoader | EllipsisType = ...,
    ) -> T:
        """Load a resource with a pydantic TypeAdapter relative to the current test."""
        data = self.load_json(*parts, ext=ext, path_maker=path_maker, json_loader=json_loader)
        adapter: TypeAdapter[T] = type_ if isinstance(type_, TypeAdapter) else TypeAdapter(type_)
        return adapter.validate_python(data)

    def save_pydantic_adapter(
        self,
        data: Any,
        *parts,
        ext: str = "json",
        path_maker: PathMaker | None = None,
        json_encoder: JsonEncoder | EllipsisType = ...,
        ndigits: int | None | EllipsisType = ...,
        type_: type[T] | TypeAdapter[T] | None = None,
        context: Any = None,
    ) -> None:
        """Write pydantic data to a resource relative to the current test."""
        type_ = type_ or type(data)
        adapter: TypeAdapter[T] = type_ if isinstance(type_, TypeAdapter) else TypeAdapter(type_)
        actual_data: T = adapter.dump_python(data, mode="json", context=context)
        self.save_json(actual_data, *parts, ext=ext, path_maker=path_maker, json_encoder=json_encoder, ndigits=ndigits)

    def delete_pydantic_adapter(self, *parts, ext: str = "json", path_maker: PathMaker | None = None):
        """Delete a json resource relative to the current test."""
        self.delete(*parts, ext=ext, path_maker=path_maker)

    def expect_pydantic_adapter(
        self,
        actual: Any,
        *parts,
        ext: str = "json",
        path_maker: PathMaker | None = None,
        json_encoder: JsonEncoder | EllipsisType = ...,
        ndigits: int | None | EllipsisType = ...,
        type_: type[T] | TypeAdapter[T] | None = None,
        context: Any = None,
    ) -> None:
        """Assert that the type adapter encodes the actual value to the JSON content from resource. This allows us to
        pass a context to the serializers of any objects embedded within actual."""
        type_ = type_ or type(actual)
        adapter: TypeAdapter[T] = type_ if isinstance(type_, TypeAdapter) else TypeAdapter(type_)
        actual_data: T = adapter.dump_python(actual, mode="json", context=context)
        self.expect_json(
            actual_data,
            *parts,
            ext=ext,
            path_maker=path_maker,
            json_encoder=json_encoder,
            ndigits=ndigits,
        )



import datetime as dt
from collections.abc import Callable, Collection, Iterable, Mapping
from functools import partial
from itertools import chain
from types import EllipsisType, UnionType
from typing import Any, Literal, TypeVar, overload

T = TypeVar("T")


@overload
def coalesce(default: T | None | None, *args: T | None | EllipsisType, nonable: Literal[True]) -> T | None: ...
@overload
def coalesce(default: T, *args: T | None | EllipsisType, nonable: Literal[False]) -> T: ...
@overload
def coalesce(default: T, *args: T | None | EllipsisType) -> T: ...


def coalesce(default: T | None, *args: T | None | EllipsisType, nonable: bool = False) -> T | None:
    """Return the first value among default, *args that is not ... and, if nonable is Fals, is not None either."""
    value: T | None = default
    for arg in args:
        if not isinstance(arg, EllipsisType) and (arg is not None or nonable is True):
            value = arg
    return value


class AbortJsonPrep(Exception):  # noqa: N818
    """Raised by a JSON prepper to indicate that even though the argument is of the expected type, it should not be
    handled by this prepper and any other ones should be given a chance."""


JSON_PREPPERS: list[tuple[type | UnionType, Callable[[Any], Any]]] = [
    (dt.date | dt.time | dt.datetime, lambda v: v.isoformat()),
]
"""List of types along with functions to prepare instances of (any sub-class of) those types for JSON encoding."""


def add_json_prepper(type_: type | UnionType, prepper: Callable[[Any], Any]) -> None:
    """Register a global JSON prepper for a given type, including sub-classes.

    The prepper can return a few kinds of values:
    - Simple value: encoded as-is and must be JSON serializable.
    - Dict: encoded recursively but must have keys that are supported by the json_encoder use, which is usually str.
    - Collection: list, tuple, set, etc will be recursively encoded as a list.

    It can also raise AbortJsonPrep to skip this prepper and continue trying others.
    """
    JSON_PREPPERS.append((type_, prepper))


# Try importing optional dependencies and register preppers for their types
try:
    from pydantic import BaseModel

    add_json_prepper(BaseModel, lambda v: v.model_dump(mode="json"))
except ImportError:
    pass

try:
    import numpy as np

    add_json_prepper(np.ndarray, lambda v: v.tolist())
    add_json_prepper(np.floating, lambda v: float(v))
except ImportError:
    pass


def prepare_for_json_encode(
    value: Any,
    *,
    ndigits: int | None = None,
    allow_negative_zero: bool = False,
    extra_preppers: Iterable[tuple[type | UnionType, Callable[[Any], Any]]] = tuple(),
) -> Any:
    """
    Copy a structure of lists, tuples, dicts, pydantic models and numpy values into a parallel structure of dicts and
    lists, trying to make them JSON encodable.

    The encoding is specifically intended for writing expectation files, so it doesn't need to be reversible, and if
    data or precision is lost in a way that is not acceptable, then the user has the opportunity to register a custom
    stringer.

    Args:
        value: The value to prepare for JSON encoding
        ndigits: The number of digits to round floats to, or None to omit rounding
        allow_negative_zero: If False, convert negative zero to plain zero in output
        extra_preppers: Additional preppers to apply before the default ones.
    """

    # Apply the configured preppers
    for type_, prepper in chain(extra_preppers, JSON_PREPPERS):
        if isinstance(value, type_):
            try:
                value = prepper(value)
                continue
            except AbortJsonPrep:
                pass

    # Apply rounding of floats values
    if isinstance(value, float):
        if ndigits is not None:
            value = round(value, ndigits)
        if not allow_negative_zero:
            value += 0.0
        return value

    # Return other JSON encodable values directly
    elif isinstance(value, str | int | bool | None):
        return value

    recurse = partial(
        prepare_for_json_encode,
        ndigits=ndigits,
        allow_negative_zero=allow_negative_zero,
        extra_preppers=extra_preppers,
    )

    # Recurse into dicts and collections
    if isinstance(value, Mapping):
        return {recurse(key): recurse(value) for key, value in value.items()}
    elif isinstance(value, Collection):
        # Collections other than strings are encoded to lists
        return [recurse(x) for x in value]

    # Convert anything else to a string
    else:
        return str(value)



from math import isnan, nan
from unittest.mock import ANY

import pytest
from pytest import FixtureRequest

from pytest_respect.ext.jsonyx import jsonyx_compactish_encoder, jsonyx_permissive_loader
from pytest_respect.resources import TestResources

# All these tests need jsonyx
pytestmark = pytest.mark.jsonyx


@pytest.fixture
def resources(request: FixtureRequest) -> TestResources:
    """The fixture being tested."""
    return TestResources(request)


def test_load_json__jsonyx_permissive(resources):
    data = resources.load_json(json_loader=jsonyx_permissive_loader)
    assert data == {"look": ["I", "found", "a", ANY]}
    assert isnan(data["look"][-1])


def test_expected_json__jsonyx(resources):
    resources.expect_json(
        {
            "look": ["what", "I", "found"],
            "numbers": [1, 2, 3, nan],
            "sub": {"simple": 1, "dict": 2, "is": 3, "flat": 4},
        },
        json_encoder=jsonyx_compactish_encoder,
    )



import datetime as dt
import json

import pytest

from pytest_respect.utils import coalesce, prepare_for_json_encode

# Optional imports, falling back to stub classes
try:
    from pydantic import BaseModel
except ImportError:
    from pytest_respect._fakes import BaseModel


def test_coalesce__given():
    assert coalesce(42, None, ..., 888, None, ...) == 888


def test_coalesce__not_given():
    assert coalesce(42, None, ...) == 42


def test_coalesce__nonable__given__not_none():
    assert coalesce(42, 888, ..., nonable=True) == 888


def test_coalesce__nonable__given__none():
    assert coalesce(42, None, ..., nonable=True) is None


@pytest.mark.parametrize("default", [42, None])
def test_coalesce__nonable__not_given(default):
    assert coalesce(default, ..., ..., nonable=True) == default


def test_prepare_for_json_encode__simple():
    import uuid

    original = [
        {
            "f-zero": 0.123456789,
            "i-zero": 0,
            "s-zero": "0",
            "stringable": uuid.NAMESPACE_X500,
        },
        [1, 2, 3.333333333],
        ("a", "b", 5.1234321),
    ]

    assert prepare_for_json_encode(original) == [
        {
            "f-zero": 0.123456789,
            "i-zero": 0,
            "s-zero": "0",
            "stringable": "6ba7b814-9dad-11d1-80b4-00c04fd430c8",
        },
        [1, 2, 3.333333333],
        ["a", "b", 5.1234321],
    ]


def test_prepare_for_json_encode__round():
    original = [
        {
            "f-zero": 0.123456789,
            "i-zero": 0,
            "s-zero": "0",
        },
        [1, 2, 3.333333333],
        ("a", "b", 5.1234321),
    ]

    assert prepare_for_json_encode(original, ndigits=4) == [
        {
            "f-zero": 0.1235,  # last digit up
            "i-zero": 0,
            "s-zero": "0",
        },
        [1, 2, 3.3333],
        ["a", "b", 5.1234],  # last digit down
    ]


def test_prepare_for_json_encode__collections():
    map = {1: "a", 2: "b", 3: "c"}
    original = {
        "frozenset": frozenset((4, 5, 6)),
        "items": map.items(),
        "keys": map.keys(),
        "set": {1, 2, 3},
        "values": map.values(),
    }

    assert prepare_for_json_encode(original) == {
        "frozenset": [4, 5, 6],
        "items": [[1, "a"], [2, "b"], [3, "c"]],
        "keys": [1, 2, 3],
        "set": [1, 2, 3],
        "values": ["a", "b", "c"],
    }


def test_prepare_for_json_encode__date_and_time():
    original = {
        "date": dt.date(2025, 7, 22),
        "time": dt.time(11, 9, 23),
        "time_utc": dt.time(11, 9, 23, tzinfo=dt.timezone.utc),
        "datetime": dt.datetime(2025, 7, 22, 11, 9, 23),
        "datetime_utc": dt.datetime(2025, 7, 22, 11, 9, 23, tzinfo=dt.timezone.utc),
    }

    prepared = prepare_for_json_encode(original)

    assert prepared == {
        "date": "2025-07-22",
        "datetime": "2025-07-22T11:09:23",
        "datetime_utc": "2025-07-22T11:09:23+00:00",
        "time": "11:09:23",
        "time_utc": "11:09:23+00:00",
    }
    json.dumps(prepared)


class PydanticModel(BaseModel):  # type: ignore (fake is insufficient)
    name: str
    weight: float
    when: dt.datetime


@pytest.mark.pydantic
def test_prepare_for_json_encode__pydantic_model():
    when = dt.datetime(1986, 3, 1, 12, 34, 56)
    original = [
        0.111111,
        PydanticModel(name="foo", weight=75.4321, when=when),
        0.555555,
    ]

    prepared = prepare_for_json_encode(original, ndigits=2)

    assert prepared == [
        0.11,
        {"name": "foo", "weight": 75.43, "when": "1986-03-01T12:34:56"},
        0.56,
    ]
    json.dumps(prepared)


@pytest.mark.pydantic
def test_prepare_for_json_encode__pydantic_model__json_mode():
    when = dt.datetime(1986, 3, 1, 12, 34, 56)
    original = [
        0.111111,
        PydanticModel(name="foo", weight=75.4321, when=when),
        0.555555,
    ]

    prepared = prepare_for_json_encode(original, ndigits=2)

    assert prepared == [
        0.11,
        {"name": "foo", "weight": 75.43, "when": "1986-03-01T12:34:56"},
        0.56,
    ]
    json.dumps(prepared)


@pytest.mark.numpy
def test_prepare_for_json_encode__numpy():
    import numpy as np

    original = [
        0.111111,
        np.arange(2, 5) * 1 / 9,  # 1-D array
        np.full((2, 3), 10 / 3),  # 2-D array
        np.full(1, 1 / 7)[0],  # scalar
        0.555555,
    ]
    assert isinstance(original[-2], np.floating)

    prepared = prepare_for_json_encode(original, ndigits=2)

    assert prepared == [
        0.11,
        [0.22, 0.33, 0.44],
        [[3.33, 3.33, 3.33], [3.33, 3.33, 3.33]],
        0.14,
        0.56,
    ]
    json.dumps(prepared)


def test_prepare_for_json_encode__allow_negative_zero__default_to_False():
    precise = [
        -0.001,
        [-0.001, -0.003, -0.005],
    ]

    # Compare json dumps since 0.0 == -0.0 as numbers
    rounded = prepare_for_json_encode(precise, ndigits=2)
    assert json.dumps(rounded) == "[0.0, [0.0, 0.0, -0.01]]"


def test_prepare_for_json_encode__allow_negative_zero__True():
    precise = [
        -0.001,
        [-0.001, -0.003, -0.005],
    ]

    # Compare json dumps since 0.0 == -0.0 as numbers
    rounded = prepare_for_json_encode(precise, ndigits=2, allow_negative_zero=True)
    assert json.dumps(rounded) == "[-0.0, [-0.0, -0.0, -0.01]]"


@pytest.mark.numpy
def test_prepare_for_json_encode__allow_negative_zero__default_to_False__numpy():
    import numpy as np

    precise = [
        -0.001,
        np.array([-0.001, -0.003, -0.005]),
    ]

    # Compare json dumps since 0.0 == -0.0 as numbers
    rounded = prepare_for_json_encode(precise, ndigits=2)
    assert json.dumps(rounded) == "[0.0, [0.0, 0.0, -0.01]]"


@pytest.mark.numpy
def test_prepare_for_json_encode__allow_negative_zero__True__numpy():
    import numpy as np

    precise = [
        -0.001,
        np.array([-0.001, -0.003, -0.005]),
    ]

    # Compare json dumps since 0.0 == -0.0 as numbers
    rounded = prepare_for_json_encode(precise, ndigits=2, allow_negative_zero=True)
    assert json.dumps(rounded) == "[-0.0, [-0.0, -0.0, -0.01]]"



from datetime import date, datetime
from pathlib import Path
from typing import Annotated
from unittest.mock import Mock, call

import pytest
from pytest import FixtureRequest
from pytest_mock import MockerFixture

from pytest_respect.resources import (
    Defaults,
    JsonLoader,
    PathMaker,
    TestResources,
    list_dir,
    list_resources,
    python_compact_json_encoder,
    python_json_encoder,
    python_json_loader,
)
from pytest_respect.utils import AbortJsonPrep

# Optional imports falling back to stub implementations to make the type checker happy
try:
    from pydantic import BaseModel, ValidationError, WrapSerializer
except ImportError:  # pragma: no cover
    from pytest_respect._fakes import BaseModel, ValidationError, WrapSerializer


THIS_FILE = Path(__file__).absolute()
THIS_DIR = THIS_FILE.parent


@pytest.fixture(scope="module", autouse=True)
def dont_tracebackhide():
    """Tests in this module want to see tracebacks from inside the resources module."""
    import pytest_respect.resources as module

    previous = module.__tracebackhide__
    module.__tracebackhide__ = False
    yield
    module.__tracebackhide__ = previous


@pytest.fixture
def mock_list_dir(mocker: MockerFixture):
    return mocker.patch(
        "pytest_respect.resources.list_dir",
        autospec=True,
        return_value=["scenario_1.json", "scenario_2.json"],
    )


@pytest.fixture
def resources(request: FixtureRequest) -> TestResources:
    """The fixture being tested."""
    return TestResources(
        request,
        accept_count=0,  # We set accept in individual tests instead of using the --respect-accept flag.
    )


@pytest.fixture
def mock_delete(resources, mocker: MockerFixture):
    """Mock resources.delete method"""
    return mocker.patch.object(resources, "delete", autospec=True)


@pytest.fixture
def resources_4digits(request: FixtureRequest) -> TestResources:
    """A TestResrouces fixture which defaults to rounding to 4 digits."""
    resources = TestResources(request)
    resources.default.ndigits = 4
    return resources


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Defaults


def test_Defaults_ndigits():
    defaults = Defaults()
    assert defaults._ndigits(...) is None
    assert defaults._ndigits(None) is None
    assert defaults._ndigits(4) == 4


def test_Defaults_json_encoder():
    defaults = Defaults()
    assert defaults._json_encoder(...) is python_json_encoder
    assert defaults._json_encoder(None) is python_json_encoder
    assert defaults._json_encoder(python_compact_json_encoder) is python_compact_json_encoder


def test_Defaults_json_loader():
    defaults = Defaults()
    assert defaults._json_loader(...) is python_json_loader
    assert defaults._json_loader(None) is python_json_loader

    mock_loader = Mock(spec=JsonLoader)
    assert defaults._json_loader(mock_loader) is mock_loader


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Static Resource Listing


def test_list_dir(tmp_path: Path):
    # This test is not very stable and needs to be updated when the dir content changes
    # Having tested the file-access code we can use `mock_list_dir` for the rest.
    listed = list_dir(THIS_DIR / "test_resources", "*_load_*", exclude="*__actual.*")
    assert listed == [
        "test_expected_json__ndigits__test_load_json.json",
        "test_load_json.json",
        "test_load_pydantic.json",
        "test_load_pydantic_adapter.json",
        "test_load_pydantic_adapter__failing.json",
        "test_load_text.txt",
    ]


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# JSON encoders & Decoders

# TODO test_python_json_encoder
# TODO test_python_compact_json_encoder
# TODO test_python_json_loader

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Path Makers & Paths


@pytest.mark.parametrize(
    "path_maker, ex_path",
    [
        [TestResources.pm_function, THIS_DIR / "test_resources__test_resources__dir"],
        [TestResources.pm_class, THIS_DIR / "test_resources"],
        [TestResources.pm_file, THIS_DIR / "test_resources"],
        [TestResources.pm_dir_named("stuff"), THIS_DIR / "stuff"],
    ],
    ids=str,
)
def test_resources__dir(resources, path_maker: PathMaker, ex_path: Path):
    """Build resource folders for different scopes."""
    path = resources.dir(path_maker)
    assert str(path) == str(ex_path)


@pytest.mark.parametrize(
    "parts, ext, ex_path",
    [
        [
            [],
            "",
            THIS_DIR / "test_resources" / "test_resources__path",
        ],
        [
            [],
            "json",
            THIS_DIR / "test_resources" / "test_resources__path.json",
        ],
        [
            ["input", "first"],
            "json",
            THIS_DIR / "test_resources" / "test_resources__path__input__first.json",
        ],
    ],
)
def test_resources__path(resources, parts: tuple[str], ext: str, ex_path):
    """Build resource paths with different parts and extension."""
    assert str(resources.path(*parts, ext=ext)) == str(ex_path)


@pytest.mark.parametrize(
    "path_maker, ex_path",
    [
        [
            TestResources.pm_function,
            THIS_DIR / "test_resources__test_resources__path__path_maker" / "data.txt",
        ],
        [
            TestResources.pm_class,
            THIS_DIR / "test_resources" / "test_resources__path__path_maker.txt",
        ],
        [
            TestResources.pm_file,
            THIS_DIR / "test_resources" / "test_resources__path__path_maker.txt",
        ],
        [
            TestResources.pm_dir,
            THIS_DIR / "resources" / "test_resources__test_resources__path__path_maker.txt",
        ],
        [
            TestResources.pm_dir_named("treasures"),
            THIS_DIR / "treasures" / "test_resources__test_resources__path__path_maker.txt",
        ],
    ],
    ids=str,
)
def test_resources__path__path_maker(resources, path_maker: PathMaker, ex_path: Path):
    """Build resource paths for different scopes."""
    path = resources.path(ext="txt", path_maker=path_maker)
    assert str(path) == str(ex_path)


class TestClass:
    @pytest.mark.parametrize(
        "path_maker, ex_path",
        [
            [
                TestResources.pm_function,
                THIS_DIR / "test_resources__TestClass__test_resources__dir",
            ],
            [TestResources.pm_class, THIS_DIR / "test_resources__TestClass"],
            [TestResources.pm_file, THIS_DIR / "test_resources"],
            [TestResources.pm_dir_named("stuff"), THIS_DIR / "stuff"],
        ],
        ids=str,
    )
    def test_resources__dir(self, resources, path_maker: PathMaker, ex_path: Path):
        """Build resource folders for different scopes."""
        path = resources.dir(path_maker)
        assert str(path) == str(ex_path)

    @pytest.mark.parametrize(
        "path_maker, ex_path",
        [
            [
                TestResources.pm_function,
                THIS_DIR / "test_resources__TestClass__test_resources__path__path_maker" / "data.txt",
            ],
            [
                TestResources.pm_class,
                THIS_DIR / "test_resources__TestClass" / "test_resources__path__path_maker.txt",
            ],
            [
                TestResources.pm_file,
                THIS_DIR / "test_resources" / "TestClass__test_resources__path__path_maker.txt",
            ],
            [
                TestResources.pm_dir,
                THIS_DIR / "resources" / "test_resources__TestClass__test_resources__path__path_maker.txt",
            ],
            [
                TestResources.pm_dir_named("treasures"),
                THIS_DIR / "treasures" / "test_resources__TestClass__test_resources__path__path_maker.txt",
            ],
        ],
        ids=str,
    )
    def test_resources__path__path_maker(self, resources, path_maker: PathMaker, ex_path: Path):
        """Build resource paths for different scopes."""
        path = resources.path(ext="txt", path_maker=path_maker)
        assert str(path) == str(ex_path)


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# List Resources


def test_list_resources__patterns(mock_list_dir):
    listed = list_resources(
        "scenario_*.json",
        exclude=("*__actual*", "*__output*", "*__aggregate"),
    )
    assert listed == mock_list_dir.return_value
    assert mock_list_dir.call_args_list == [
        call(
            THIS_DIR / "test_resources",
            "scenario_*.json",
            exclude=("*__actual*", "*__output*", "*__aggregate"),
        )
    ]


def test_list_resources__path_maker(mock_list_dir):
    listed = list_resources(path_maker=TestResources.pm_only_dir)
    assert listed == mock_list_dir.return_value
    assert mock_list_dir.call_args_list == [
        call(
            THIS_DIR / "resources",
            "*",
            exclude=tuple(),
        )
    ]


def test_list_resources__strip_ext_True(mock_list_dir):
    mock_list_dir.return_value = [
        "resource_1.json",
        "resource_2.json",
        "resource_3.txt",
    ]
    listed = list_resources(strip_ext=True)
    assert listed == [
        "resource_1",
        "resource_2",
        "resource_3",
    ]


def test_list_resources__strip_ext_json(mock_list_dir):
    mock_list_dir.return_value = [
        "resource_1.json",
        "resource_2.json",
        "resource_3.txt",
    ]
    listed = list_resources(strip_ext=".json")
    assert listed == [
        "resource_1",
        "resource_2",
        "resource_3.txt",
    ]


@pytest.fixture(params=list_resources("*.json", exclude=["*__actual*"], strip_ext=True))
def each_resource_name(request) -> str:
    """Dependants run for name of each file in the resources folder."""
    return request.param


def test_list_resources(resources, each_resource_name):
    # The resources names already include the full file name
    print(each_resource_name)
    resources.default.path_maker = resources.pm_only_file
    path = resources.path(each_resource_name, ext="json")
    assert path.is_file()
    resources.load_json(each_resource_name)


def test_list__defaults(resources, mock_list_dir):
    assert resources.list() == mock_list_dir.return_value
    assert mock_list_dir.call_args_list == [
        call(
            THIS_DIR / "test_resources",
            "*",
            exclude=tuple(),
            strip_ext=False,
        )
    ]


def test_list__path_maker(resources, mock_list_dir):
    listed = resources.list(path_maker=TestResources.pm_only_dir_named("stuff"))
    assert listed == mock_list_dir.return_value
    assert mock_list_dir.call_args_list == [
        call(
            THIS_DIR / "stuff",
            "*",
            exclude=tuple(),
            strip_ext=False,
        )
    ]


def test_list__filtered(resources, mock_list_dir):
    listed = resources.list("*_load_*", exclude="*__actual.*")
    assert listed == mock_list_dir.return_value

    assert mock_list_dir.call_args_list == [
        call(
            THIS_DIR / "test_resources",
            "*_load_*",
            exclude="*__actual.*",
            strip_ext=False,
        )
    ]


def test_delete(resources):
    path = resources.path()
    path.write_text("temporary")
    assert path.is_file()

    resources.delete()
    assert not path.is_file()


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Text Resources


def test_load_text(resources):
    text = resources.load_text()
    assert text == "text resource\n"


def test_save_text__dir_does_not_exist(resources):
    """We can write a file to a directory that doesn't exist yet."""
    my_dir = Path(__file__).parent
    missing_dir_name = "missing_dir"
    missing_dir = my_dir / missing_dir_name
    new_file_name = "new_file.txt"
    new_file = missing_dir / new_file_name

    try:
        # Clean up file and dir from previous run. If other files have been added, remove manually.
        new_file.unlink(missing_ok=True)
        if missing_dir.is_dir():
            missing_dir.rmdir()

        def make_missing_path(*a, **kw):
            return (missing_dir, new_file_name)

        resources.save_text("Some text is here", path_maker=make_missing_path)

        text = new_file.read_text()
        assert text == "Some text is here"

    finally:
        new_file.unlink(missing_ok=True)
        if missing_dir.is_dir():
            missing_dir.rmdir()


def test_delete_text(resources, mock_delete):
    resources.delete_text("one", "two", path_maker=resources.pm_file)

    mock_delete.assert_called_once_with("one", "two", ext="txt", path_maker=resources.pm_file)


@pytest.mark.parametrize("accept", [1, 0])
def test_expect_text__match(resources, accept: bool):
    resources.accept_count = accept
    resources.expect_text("some text\nsome more text\n")


def test_expect_text__mismatch(resources):
    try:
        resources.delete_text("actual")
        with pytest.raises(AssertionError):
            resources.expect_text("actual text not found")

        # Actual file was written
        actual_path = resources.path("actual", ext="txt")
        assert actual_path.exists()
        written_actual = resources.load_text("actual")
        assert written_actual == "actual text not found"

        # Expected file was not changed
        resources.expect_text("original expected test\n")

    finally:
        resources.delete_text("actual")


def test_expect_text__mismatch__accept(resources, capsys):
    resources.accept_count = 1
    try:
        resources.save_text("previous actual\n", "actual")
        resources.save_text("original expected test\n")

        resources.expect_text("updated expected test\n")

        # The actual file was removed
        actual_path = resources.path("actual", ext="txt")
        assert not actual_path.exists()

        # The expected file was overwritten
        written_expected = resources.load_text()
        assert written_expected == "updated expected test\n"

        expected_path = resources.path(ext="txt")
        assert capsys.readouterr().out.splitlines()[-1] == f"The expectation file was updated at {expected_path}."

    finally:
        resources.delete_text()
        resources.delete_text("actual")


def test_expect_text__not_found(resources):
    test_dir = Path(__file__).with_suffix("")
    expected_file = test_dir / "test_expect_text__not_found.txt"
    actual_file = test_dir / "test_expect_text__not_found__actual.txt"
    text_content = "actual text not found"

    expected_file.unlink(missing_ok=True)
    actual_file.unlink(missing_ok=True)

    try:
        with pytest.raises(AssertionError) as exi:
            resources.expect_text(text_content)

        assert str(exi.value).startswith("The expectation file was not found at ")
        assert "test_expect_text__not_found.txt" in str(exi.value)

        assert not expected_file.exists()
        assert actual_file.exists()
        assert actual_file.read_text() == text_content

    finally:
        expected_file.unlink(missing_ok=True)
        actual_file.unlink(missing_ok=True)


def test_expect_text__not_found__accept(resources, capsys):
    resources.accept_count = 1

    test_dir = Path(__file__).with_suffix("")
    expected_file = test_dir / "test_expect_text__not_found__accept.txt"
    actual_file = test_dir / "test_expect_text__not_found__accept__actual.txt"
    text_content = "actual text not found"

    expected_file.unlink(missing_ok=True)
    actual_file.write_text("previous actual\n")

    try:
        resources.expect_text(text_content)

        assert expected_file.exists()
        assert expected_file.read_text() == text_content

        assert not actual_file.exists()

        assert capsys.readouterr().out.splitlines()[-1] == f"A new expectation file was written to {expected_file}."
    finally:
        expected_file.unlink(missing_ok=True)
        actual_file.unlink(missing_ok=True)


def test_accept_count(resources):
    assert resources.accept_count == 0
    resources.accept_count = 2
    resources.delete_text()  # Start with no expectation and accept 2 mismatches

    # Allow one failure
    resources.expect_text("first value")
    assert resources.accept_count == 1

    # Allow another failure
    resources.expect_text("second value")
    assert resources.accept_count == 0

    # Don't allow a third failure
    with pytest.raises(AssertionError):
        resources.expect_text("third value")

    # Second value is still in file
    resources.expect_text("second value")

    resources.delete_text()


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# JSON Resources


def test_data_to_json__add_json_prepper(resources):
    class CustomType:
        def __init__(self, name: str):
            self.name = name

    data = {
        1: "something",
        2: CustomType("thingy"),
    }

    resources.add_json_prepper(CustomType, lambda ct: "my custom " + ct.name)

    resources.default.json_encoder = python_compact_json_encoder
    assert resources.data_to_json(data) == '{"1": "something", "2": "my custom thingy"}\n'


def test_data_to_json__abort_json_prepper(resources):
    class CustomType:
        def __init__(self, name: str):
            self.name = name

        def __str__(self):
            return f"CustomType({self.name})"

    aborting_prepper_called = False

    def aborting_prepper(ct: CustomType):
        nonlocal aborting_prepper_called
        aborting_prepper_called = True
        raise AbortJsonPrep()

    data = {
        1: "something",
        2: CustomType("thingy"),
    }

    resources.add_json_prepper(CustomType, aborting_prepper)

    resources.default.json_encoder = python_compact_json_encoder
    assert resources.data_to_json(data) == '{"1": "something", "2": "CustomType(thingy)"}\n'

    assert aborting_prepper_called


def test_load_json(resources):
    data = resources.load_json()
    assert data == {"look": ["what", "I", "found"]}


def test_load_json__missing(resources):
    with pytest.raises(ValueError) as exi:
        resources.load_json()

    assert str(exi.value).startswith("Failed to load JSON resource")
    assert "pytest_resources/test_resources/test_load_json__missing.json" in str(exi.value)


def test_load_json__overrides(resources):
    json_loader = Mock(wraps=resources.default.json_loader)

    data = resources.load_json(
        "test_load_json",
        path_maker=resources.pm_only_file,  # Override allows us to find the file
        json_loader=json_loader,
    )

    assert data == {"look": ["what", "I", "found"]}
    assert json_loader.call_count == 1


def test_save_json(resources):
    resources.delete_json()
    data = {
        "here": ["is", "a", "new", "one"],
        "foo": 1.23456,
    }

    resources.save_json(data)

    assert resources.load_json() == {
        "here": ["is", "a", "new", "one"],
        "foo": 1.23456,
    }
    resources.delete_json()


def test_save_json__overrides(resources):
    resources.delete_json()

    json_encoder = Mock(wraps=resources.default.json_encoder)
    data = {
        "here": ["is", "a", "new", "one"],
        "foo": 1.23456,
    }

    resources.save_json(
        data,
        json_encoder=json_encoder,
        ndigits=2,
    )

    assert resources.load_json() == {
        "here": ["is", "a", "new", "one"],
        "foo": 1.23,
    }
    assert json_encoder.call_count == 1
    resources.delete_json()


def test_delete_json(resources, mock_delete):
    resources.delete_json("one", "two", path_maker=resources.pm_file)

    mock_delete.assert_called_once_with("one", "two", ext="json", path_maker=resources.pm_file)


def test_expected_json(resources):
    resources.expect_json(
        {
            "look": ["what", "I", "found"],
            "float": 0.1234,
        }
    )


def test_expected_json__compact(resources):
    resources.default.json_encoder = python_compact_json_encoder
    resources.expect_json(
        {
            "look": ["what", "I", "found"],
            "float": 0.1234,
        }
    )


def test_expected_json__ndigits(resources):
    actual = {
        "float": 0.1234321,  # more digits than the resource
    }

    with pytest.raises(AssertionError):
        resources.expect_json(actual, "test_load_json")

    with pytest.raises(AssertionError):
        resources.expect_json(actual, "test_load_json", ndigits=6)

    resources.expect_json(actual, "test_load_json", ndigits=4)


def test_expected_json__default_digits(resources_4digits):
    actual = {
        "float": 0.1234321,  # more digits than the resource
    }

    assert resources_4digits.default.ndigits == 4
    resources_4digits.expect_json(actual)


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Pydantic Resources


class MyModel(BaseModel):  # type: ignore
    look: list[str | float]


class MyModelWithDates(BaseModel):  # type: ignore
    date: date
    datetime: datetime


@pytest.mark.pydantic
def test_load_pydantic(resources):
    data = resources.load_pydantic(MyModel)
    assert data == MyModel(look=["I", "found", "this"])


@pytest.mark.pydantic
def test_load_pydantic__overrides(resources):
    json_loader = Mock(wraps=resources.default.json_loader)

    data = resources.load_pydantic(
        MyModel,
        "test_load_pydantic",
        path_maker=resources.pm_only_file,  # Override allows us to find the file
        json_loader=json_loader,
    )

    assert data == MyModel(look=["I", "found", "this"])
    assert json_loader.call_count == 1


def add_context(x, handler, info) -> list[str]:
    x = handler(x)
    if info.context:
        return x + info.context
    return x


class MyModelWithContext(BaseModel):  # type: ignore
    look: Annotated[
        list[str],
        WrapSerializer(add_context),
    ]
    """A property whose custom serializer adds to it whatever is in the serialization context."""


@pytest.mark.pydantic
def test_save_pydantic(resources):
    resources.delete_pydantic()
    data = MyModel(look=["saved", "pydantic", "data", 1.2345])

    resources.save_pydantic(data)

    assert resources.load_json() == {
        "look": ["saved", "pydantic", "data", 1.2345],
    }
    resources.delete_pydantic()


@pytest.mark.pydantic
def test_save_pydantic__with_context(resources):
    resources.delete_pydantic()
    data = MyModelWithContext(look=["saved", "pydantic", "data"])

    resources.save_pydantic(data, ndigits=2, context=["with", "context"])

    assert resources.load_json() == {
        "look": ["saved", "pydantic", "data", "with", "context"],
    }
    resources.delete_pydantic()


@pytest.mark.pydantic
def test_save_pydantic__overrides(resources):
    resources.delete_pydantic()

    json_encoder = Mock(wraps=resources.default.json_encoder)
    data = MyModel(look=["saved", "pydantic", "data", 1.2345])

    resources.save_pydantic(data, json_encoder=json_encoder, ndigits=2)

    assert resources.load_json() == {
        "look": ["saved", "pydantic", "data", 1.23],
    }
    resources.delete_pydantic()

    assert json_encoder.call_count == 1


def test_delete_pydantic(resources, mock_delete):
    resources.delete_pydantic("one", "two", path_maker=resources.pm_file)

    mock_delete.assert_called_once_with("one", "two", ext="json", path_maker=resources.pm_file)


@pytest.mark.pydantic
def test_expected_pydantic(resources):
    resources.expect_pydantic(MyModel(look=["I", "was", "expecting", "this"]))


@pytest.mark.pydantic
def test_expected_pydantic__with_context(resources):
    resources.expect_pydantic(
        MyModelWithContext(look=["I", "was", "expecting", "this"]),
        context=["with", "context"],
    )

    # Show that the expectation contains the context
    assert resources.load_json()["look"] == ["I", "was", "expecting", "this", "with", "context"]


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Pydantic TypeAdapter Resources


@pytest.mark.pydantic
def test_load_pydantic_adapter(resources):
    data = resources.load_pydantic_adapter(dict[str, int])
    assert data == {"a": 1, "b": 2, "c": 3}


@pytest.mark.pydantic
def test_load_pydantic_adapter__failing(resources):
    with pytest.raises(ValidationError) as exi:
        resources.load_pydantic_adapter(dict[str, int])

    assert exi.value.errors()[0]["msg"] == ("Input should be a valid integer, unable to parse string as an integer")


@pytest.mark.pydantic
def test_load_pydantic_adapter__overrides(resources):
    json_loader = Mock(wraps=resources.default.json_loader)

    data = resources.load_pydantic_adapter(
        dict[str, int],
        "test_load_pydantic_adapter",
        path_maker=resources.pm_only_file,  # Override allows us to find the file
        json_loader=json_loader,
    )

    assert data == {"a": 1, "b": 2, "c": 3}
    assert json_loader.call_count == 1


@pytest.mark.pydantic
def test_save_pydantic_adapter(resources):
    resources.delete_pydantic()
    data: dict[int, MyModelWithContext] = {
        1: MyModelWithContext(look=["I", "was", "expecting", "this"]),
        2: MyModel(look=["and", "also", "this", 1.2345]),  # Won't have the context
    }

    resources.save_pydantic_adapter(data, context=["with", "context"])

    assert resources.load_json() == {
        "1": {"look": ["I", "was", "expecting", "this", "with", "context"]},
        "2": {"look": ["and", "also", "this", 1.2345]},  # without context
    }
    resources.delete_pydantic()


@pytest.mark.pydantic
def test_save_pydantic_adapter__overrides(resources):
    resources.delete_pydantic()

    json_encoder = Mock(wraps=resources.default.json_encoder)
    data: dict[int, MyModelWithContext] = {
        1: MyModelWithContext(look=["I", "was", "expecting", "this"]),
        2: MyModel(look=["and", "also", "this", 1.2345]),  # Won't have the context
    }

    resources.save_pydantic_adapter(
        data,
        json_encoder=json_encoder,
        ndigits=2,
        context=["with", "context"],
    )

    assert resources.load_json() == {
        "1": {"look": ["I", "was", "expecting", "this", "with", "context"]},
        "2": {"look": ["and", "also", "this", 1.23]},  # without context
    }
    assert json_encoder.call_count == 1

    resources.delete_pydantic()


def test_delete_pydantic_adapter(resources, mock_delete):
    resources.delete_pydantic("one", "two", path_maker=resources.pm_file)

    mock_delete.assert_called_once_with("one", "two", ext="json", path_maker=resources.pm_file)


@pytest.mark.pydantic
def test_expected_pydantic_adapter(resources):
    data: dict[int, MyModelWithContext] = {
        1: MyModelWithContext(look=["I", "was", "expecting", "this"]),
        2: MyModel(look=["and", "also", "this"]),  # Won't have the context
    }
    resources.expect_pydantic_adapter(data, context=["with", "context"])



"""Fake classes and other constructs with just enough stub code to make pyright accept our optional imports."""

from typing import Any, Generic, NoReturn, TypeVar, final

T = TypeVar("T")


def needs(name: str) -> NoReturn:
    raise NotImplementedError(f"This part of pytest-respect requires the optional {name} dependency.")


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Fake pydantic types


class BaseModel:
    def __init__(self, type: Any, **kw) -> NoReturn:
        needs("pydantic")

    @classmethod
    def model_validate(cls, obj: Any, *a, **kw) -> NoReturn:
        needs("pydantic")

    def model_dump(self, **kw) -> NoReturn:
        needs("pydantic")


@final
class TypeAdapter(Generic[T]):
    def __init__(self, type: Any, **kw) -> NoReturn:
        needs("pydantic")

    def validate_python(self, obj: Any, *a, **kw) -> NoReturn:
        needs("pydantic")

    def dump_python(self, instance: T, /, **kw) -> Any:
        needs("pydantic")


class ValidationError(Exception):
    def errors(self, *a, **kw) -> NoReturn:
        needs("pydantic")


class WrapSerializer:
    def __init__(self, *a, **kw):
        super().__init__()


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Fake numpy types


class ndarray:  # noqa N801
    def tolist(self) -> NoReturn:
        needs("numpy")


class ndfloat:  # noqa N801
    def __float__(self) -> NoReturn:
        needs("numpy")
